# Copyright (C) 2007-2009 Hardaker
# License: GNU GPLv2.  See the COPYING file for details.
package GeoDB::Export::Kml;

use strict;
use GeoDB::Export;
use GeoDB::Utils;
use GeoDB::Export::Density;
use IO::File;
use QWizard;
use Encode;

our $VERSION = "1.11";

our @ISA = qw(GeoDB::Export GeoDB::Export::Density);

our @altitudeValues =
  (relativeToGround => 'relative to the ground',
   absolute => 'absolute height above sea level');

our @GEOOPTIONS =
  (
   { name => 'bogus',
     type => 'label',
     text => 'Basic Settings:',
   },
   { name => 'dataname',
     text => 'Name to call the results:',
     default => 'Export from GeoQO',
     indent => 1,
   },
   { name => 'dowaypoints',
     text => 'Include waypoint/geocache markers:',
     type => 'checkbox',
     refresh_on_change => 1,
     default => 1,
   },
   { name => 'dodensity',
     text => 'Create a density map:',
     type => 'checkbox',
     refresh_on_change => 1,
     default => 1,
   },
   { name => 'dojail',
     text => 'Show .1 mile radius circles',
     type => 'checkbox',
     refresh_on_change => 1,
     default => 0,
   },
   "",
   { name => 'bogus',
     type => 'label',
     text => 'Configuration of the Waypoints:',
     doif => sub {qwparam('dowaypoints') eq '1'},
   },
   { name => 'waypointstyle',
     type => 'menu',
     text => 'Waypoint Style',
     labels => ['geoqo' => 'Unique icons per difficulty and waypoint type',
		'pushpin' => 'A single pushpin at each waypoint',
		],
     default => 'geoqo',
     indent => 1,
     doif => sub {qwparam('dowaypoints') eq '1'},
   },
   { name => 'waypointheight',
     text => 'Waypoint Altitude',
     default => 150,
     advanced => 1,
     indent => 1,
     helpdesc => '(0 means on the ground)',
     doif => sub {qwparam('dowaypoints') eq '1'},
   },
   { name => 'waypointmode',
     text => 'Altitude mode',
     type => 'menu',
     advanced => 1,
     labels => \@altitudeValues,
     default => 'relativeToGround',
     doif => sub {qwparam('dowaypoints') eq '1'},
     indent => 1,
   },
   { name => 'waypointextrude',
     type => 'checkbox',
     text => 'Draw a line to the ground:',
     advanced => 1,
     indent => 1,
     default => 1,
     doif => sub {qwparam('dowaypoints') eq '1'},
   },
   { name => 'includedescriptions',
     type => 'checkbox',
     text => 'Include cache descriptions in output:',
     indent => 1,
     default => 1,
     doif => sub {qwparam('dowaypoints') eq '1'},
   },
   "",
   { name => 'bogus',
     type => 'label',
     text => 'Configuration of the density grid:',
     doif => sub {qwparam('dodensity') eq '1'},
   },
   { name => 'squareratio',
     default => '1',
     type => 'checkbox',
     text => 'Use a square grid',
     indent => 1,
     advanced => 1,
     refresh_on_change => 1,
     doif => sub {qwparam('dodensity') eq '1'},
   },
   { name => 'size',
     default => 200,
     indent => 1,
     title => 'Number of squares on a side of the density plot',
     help => 'Creates a plot where there are SIZE by SIZE squares of colored rectangles',
     doif => sub {qwparam('squareratio') eq '1' && qwparam('dodensity') eq '1'},
   },
   { name => 'width',
     indent => 1,
     advanced => 1,
     title => 'Width in squares of the rectangular density plot',
     help => 'Normally equal to the size value but can be set independently.  It specifies the width (in squares) of the density plot.',
     doif => sub {qwparam('squareratio') ne '1' && qwparam('dodensity') eq '1'},
   },
   { name => 'height',
     indent => 1,
     advanced => 1,
     title => 'Height in squares of the rectangular density plot',
     help => 'Normally equal to the size value but can be set independently.  It specifies the height (in squares) of the density plot.',
     doif => sub {qwparam('squareratio') ne '1' && qwparam('dodensity') eq '1'},
   },
   { name => 'spread',
     title => 'Spreading Distance',
     indent => 1,
     advanced => 1,
     help => 'Sets the spread that determines the number of neighboring squares that a given cache will effect.  EG, a spread of 5 will affect a circle of density squares in a radius of 5.  This generally should be a low value of probably not more than 5 or so.',
     default => 5,
     doif => sub {qwparam('dodensity') eq '1'},
   },
   { name => 'doempty',
     title => 'Include empty squares in the density plot:',
     help => 'If set to 1, even empty squares will be colored (red).  Otherwise the empty squares are removed from the plot leaving the plane ground underneath.',
     indent => 1,
     type => 'checkbox',
     default => 0,
     doif => sub {qwparam('dodensity') eq '1'},
   },


   "",
   { name => 'bogus',
     type => 'label',
     text => 'Configuration Parameters for shape style:',
     advanced => 1,
     doif => sub {qwparam('dodensity') eq '1'},
   },
   { name => 'altitude',
     title => 'Altitude of density plot squares',
     indent => 1,
     help => 'The altitude to set the density plot squares at.',
     advanced => 1,
     default => 1000,
     doif => sub {qwparam('dodensity') eq '1'},
   },
   { name => 'densitymode',
     text => 'Altitude mode:',
     type => 'menu',
     labels => \@altitudeValues,
     default => 'relativeToGround',
     advanced => 1,
     doif => sub {qwparam('dodensity') eq '1'},
     indent => 1,
   },
   { name => 'extrude',
     title => 'Extrude the density plot to the ground:',
     help => 'Whether or not to extrude the density plot squares down to the ground.',
     type => 'checkbox',
     advanced => 1,
     indent => 1,
     default => 0,
     doif => sub {qwparam('dodensity') eq '1'},
   },
   { name => 'opaque',
     title => 'Opaqueness of the density squares:',
     helpdesc => '(0-100; 0 being completely see-through)',
     help => 'The opacity of the density squares.  Must be between 0 and 100, with 0 being completely see-through (and thus completely useless).',
     indent => 1,
     advanced => 1,
     default => 20,
     doif => sub {qwparam('dodensity') eq '1'},
   },
   { name => 'linewidth',
     title => 'Line width:',
     help => 'The width of the border lines to draw (0-4).  0 means don\'t draw borders on the squares.',
     advanced => 1,
     indent => 1,
     default => 0,
     doif => sub {qwparam('dodensity') eq '1'},
   },

   "",
   { name => 'bagus',
     type => 'label',
     text => 'Configure .1 mile radius circles:',
     advanced => 1,
     doif => sub {qwparam('dojail') eq '1'},
   },
   { name => 'jailaltitude',
     text => 'Altitude of the circles:',
     advanced => 1,
     default => '100',
     doif => sub {qwparam('dojail') eq '1'},
     indent => 1,
   },
   { name => 'jailmode',
     text => 'Altitude mode:',
     type => 'menu',
     labels => \@altitudeValues,
     advanced => 1,
     default => 'relativeToGround',
     doif => sub {qwparam('dojail') eq '1'},
     indent => 1,
   },


   "",
   { name => 'bogus',
     type => 'label',
     text => 'Limit the input data:',
   },
   { name => 'limitdata',
     default => '1',
     type => 'checkbox',
     indent => 1,
     text => 'Use all the data from the search set',
     refresh_on_change => 1,
   },
   { name => 'nmax',
     title => 'Maximum North coordinate:',
     indent => 1,
     help => 'The maximum north coordinate value to created the grid over.  Normally this is automatically set by the maximum value found in the data, but can be overridden',
     doif => sub {qwparam('limitdata') ne '1'},
   },
   { name => 'nmin',
     title => 'Minimum North coordinate:',
     indent => 1,
     help => 'The minimum north coordinate value to created the grid over.  Normally this is automatically set by the minimum value found in the data, but can be overridden',
     doif => sub {qwparam('limitdata') ne '1'},
   },
   { name => 'wmax',
     title => 'Maximum West coordinate:',
     indent => 1,
     help => 'The maximum west coordinate value to created the grid over.  Normally this is automatically set by the maximum value found in the data, but can be overridden',
     doif => sub {qwparam('limitdata') ne '1'},
   },
   { name => 'wmin',
     title => 'Minimum West coordinate:',
     indent => 1,
     help => 'The minimum west coordinate value to created the grid over.  Normally this is automatically set by the minimum value found in the data, but can be overridden',
     doif => sub {qwparam('limitdata') ne '1'},
   }
  );

sub _x {
    return encode('utf8',GeoDB::Export::_add_basic_entity(@_));
}

sub export_it {
    my ($self, $file, $set) = @_;

    GEODEBUG(1,"writing file: $file\n");
    my $fh = new IO::File ">$file";

    # start the header
    # <bounds minlat="38.236517" minlon="-122.04275" maxlat="38.870533" maxlon="-121.271167" />
    print $fh "<?xml version=\"1.0\" encoding=\"utf-8\"?>
<kml xmlns=\"http://earth.google.com/kml/2.2\">
<Document>
  <name>$self->{'options'}{'dataname'}</name>
  <open>1</open>
  <description>This file was created using GeoQO, which is available from http://www.geoqo.org/</description>
";

    # these match dist/makeicons
    my @shapes = ('micro', 'small', 'regular', 'large', 'nosize', 'multi',
		  'virtual', 'unknown', 'solved', 'webcam', 'event', 'earth');
    my @diffs = (qw(1 1.5 2 2.5 3 3.5 4 4.5 5));

    my %shapes;

    # define the styles needed
    if ($self->{'options'}{'waypointstyle'} eq 'geoqo') {
	foreach my $shape (@shapes) {
	    $shapes{$shape} = 1; # cache for later check if we can do it
	    foreach my $diff (@diffs) {
		$diff =~ s/\.//;
		print $fh "  <Style id=\"$shape$diff\">
    <IconStyle>
      <Icon>
        <href>http://www.geoqo.org/images/icons/geocaches-$shape-$diff.png</href>
      </Icon>
    </IconStyle>
  </Style>
";
	    }
	}
    }

    print $fh "<Folder>
  <name>$self->{'options'}{'dataname'}</name>
  <description>$self->{'options'}{'dataname'}</description>\n";
    # load data
    $set->load_parts({logs => 1});

    # remember options
    my $geoqostyle = 0;
    $geoqostyle = 1 if ($self->{'options'}{'waypointstyle'} eq 'geoqo');
    my $doextrude = "";
    my $doextrude = "\n      <extrude>1</extrude>"
      if ($self->{'options'}{'waypointextrude'});
    my $height = $self->{'options'}{'waypointheight'};
    my $mode = $self->{'options'}{'waypointmode'};

    my $dodescriptions = $self->{'options'}{'includedescriptions'};

    # nopoints is historic...
    if ($self->{'options'}{'dowaypoints'} && !$self->{'options'}{'nopoints'}) {

	print $fh "  <Folder>
    <name>$self->{'options'}{'dataname'}: Waypoints</name>\n";

	my $count = 0;
	my $next = 0;
	$self->setup_progress($set->num(), "Creating Waypoints");
	$set->foreach(sub {
			  $next = $self->set_progress($count++)
			    if ($count >= $next);
			  my $style = ($geoqostyle) ? get_tomtom_style($_[0])
			    : "#khStyle652";
			  my $c=$_[0]{'data'};
			  my $type = ($c->{'subtype'} ? "$c->{'type'}|$c->{'subtype'}" : $c->{'type'});

			  my $description = "";
			  $description = "<description><![CDATA[<a href=\"http://www.geocaching.com/seek/cache_details.aspx?wp=$c->{ident}\">$c->{ident} Details</a>";
			  if ($dodescriptions) {
			      $description .= "<br />";
			      if ($c->{'groundspeak_short_description_html'} eq 'True') {
				  $description .=
				    $c->{'groundspeak_short_description'};
			      } else {
				  $description .=
				    "<pre>" .
				      _x($c->{'groundspeak_short_description'}) .
					"</pre><br />";
			      }

			      if ($c->{'groundspeak_long_description_html'} eq 'True') {
				  $description .=
				    $c->{'groundspeak_long_description'}
			      } else {
				  $description .=
				    "<pre>" .
				      _x($c->{'groundspeak_long_description'}) .
					"</pre>";
			      }
			  }
			  $description .= "]]></description>";

			  print $fh "
  <Placemark>
    <name>" . _x($c->{'ident'} . ": " . $c->{'urlname'}) . "</name>$description
    <styleUrl>$style</styleUrl>
    <Point>$doextrude
      <tessellate>0</tessellate>
      <altitudeMode>$mode</altitudeMode>
      <coordinates>" . _x($c->{'lon'}) . "," . _x($c->{'lat'}) . ",$height</coordinates>
    </Point>
  </Placemark>
";
		      });
	$self->stop_progress();
	
	print $fh "</Folder>\n";
    }

    # create jails around the cache
    if ($self->{'options'}{'dojail'}) {

	print $fh "  <Folder>
    <name>$self->{'options'}{'dataname'}: Waypoint .1 Mile Boundaries</name>\n";

	# calculate a ring
	my (@distringx, @distringy);
	my $ringcount = 16;
	my $radiusdist = .001428;  # right for lat, wrong for long
	for (my $i = 0; $i < $ringcount; $i++) {
	    push @distringx, cos(2 * 3.1415 * $i / $ringcount) * $radiusdist;
	    push @distringy, sin(2 * 3.1415 * $i / $ringcount) * $radiusdist;
	}
	push @distringx, cos(0) * $radiusdist;
	push @distringy, sin(0) * $radiusdist;

	my $altitude = $self->{'options'}{'jailaltitude'};
	my $mode = $self->{'options'}{'jailmode'};
	my $color = "800000ff";
	my $linewidth = 1;
	my $extrude = 1;

	my $count = 0;
	my $next = 0;
	$self->setup_progress($set->num(), "Creating Jail Circles");
	$set->foreach(sub {
			  $count++;
			  $next = $self->set_progress($count++)
			    if ($count >= $next);
			  my $c=$_[0]{'data'};
			  print $fh "
   <Placemark id=\"jail$count\">
    <name>GeoJail$count</name>
    <Style>
      <PolyStyle>
        <color>$color</color>
      </PolyStyle>
      <LineStyle>
        <width>$linewidth</width>
        <color>$color</color>
      </LineStyle>
    </Style>
    <Polygon id=\"poly$count\">
      <extrude>$extrude</extrude>
      <tessellate>0</tessellate>
      <altitudeMode>$mode</altitudeMode>
      <outerBoundaryIs>
        <LinearRing>
          <coordinates>
";
			  for (my $i = 0; $i <= $ringcount; $i++) {
			      print $fh ($c->{'lon'}+$distringy[$i]) . "," .
				($c->{'lat'}+$distringx[$i]) . "," .
				  $altitude, " ";
			  }
			  print $fh "
          </coordinates>
        </LinearRing>
      </outerBoundaryIs>
    </Polygon>
  </Placemark>
";
		      });
	$self->stop_progress();
	
	print $fh "</Folder>\n";
    }

    # nodensity is historic
    $self->get_density_map_xml($set, $fh)
      if ($self->{'options'}{'dodensity'} && !$self->{'options'}{'nodensity'});

print $fh "</Folder>
</Document>
</kml>
";
}

my %containers =
  (
   'Micro' 	=> 'micro',
   'Small' 	=> 'small',
   'Regular'    => 'regular',
   'Not chosen' => 'nosize',
   'Other'      => 'nosize',
   'Large'      => 'large',
  );

my %types =
  (
   'Multi-cache' => 'multi',
   'Unknown Cache' => 'unknown',
   'Virtual Cache' => 'virtual',
   'Webcam Cache' => 'webcam',
   'Event Cache' => 'event',
   'Earthcache' => 'earth',
   # XXX!!!  need letterbox!
   # 'Letterbox Hybrid' => 'letter',
  );

sub get_density_map_xml {
    my ($self, $set, $fh) = @_;

    my ($spreadresults, $maxcount,
	$maxlat, $minlat, $maxlon, $minlon) = $self->get_density_map($set);
    my ($binnumx, $binnumy) = ($#$spreadresults, $#{$spreadresults->[0]});

    my $latdiff = $maxlat - $minlat;
    my $londiff = $maxlon - $minlon;

#    use Image::Imlib2;
#     my $image = Image::Imlib2->new($binnumx+1, $binnumy+1);
#     $image->set_color(255,255,255,255);
#     $image->draw_rectangle(0, 0, $binnumx+1, $binnumy+1);
#     $image->has_alpha(1);

    GEODEBUG(5, "Finally writting out results\n");
    my $dump = $self->{'options'}{'dump'};
    my $alt = $self->{'options'}{'altitude'};
    my $extrude = $self->{'options'}{'extrude'};
    my $opaque = $self->{'options'}{'opaque'};
    $opaque = int(255*$opaque/100);
    my $linewidth = $self->{'options'}{'linewidth'};
    my $dozeros = $self->{'options'}{'doempty'};
    my $name = $self->{'options'}{'dataname'};

    my $count;
    my $mode = $self->{'options'}{'densitymode'};

    print $fh "<Folder><name>$name: Density Plot</name>\n";
    for (my $i = 0; $i <= $binnumx; $i++) {
	for (my $j = 0; $j <= $binnumy; $j++) {
	    $count++;
	    if ($spreadresults->[$i][$j] > 0 || $dozeros) {
		my $color = sprintf("%02x%02x%02x%02x", 
				    reverse(hsv_to_rgb((($maxcount == 0) ? 0 : .8*$spreadresults->[$i][$j] / $maxcount)), $opaque));
		print $fh "
   <Placemark id=\"GeoCD$count\">
    <name>GeoCD$count</name>
    <Style>
      <PolyStyle>
        <color>$color</color>
      </PolyStyle>
      <LineStyle>
        <width>$linewidth</width>
        <color>$color</color>
      </LineStyle>
    </Style>
    <Polygon id=\"poly$count\">
      <extrude>$extrude</extrude>
      <tessellate>1</tessellate>
      <altitudeMode>$mode</altitudeMode>
      <outerBoundaryIs>
        <LinearRing>
          <coordinates>
";
		print $fh
		  ($minlon+$j/$binnumy*$londiff , "," ,
		   $minlat+$i/$binnumx*$latdiff , ",$alt " , 
		   $minlon+($j+1)/$binnumy*$londiff , "," ,
		   $minlat+$i/$binnumx*$latdiff , ",$alt " , 
		   $minlon+($j+1)/$binnumy*$londiff , "," ,
		   $minlat+($i+1)/$binnumx*$latdiff , ",$alt " , 
		   $minlon+$j/$binnumy*$londiff , "," ,
		   $minlat+($i+1)/$binnumx*$latdiff , ",$alt " , 
		   $minlon+$j/$binnumy*$londiff , "," ,
		   $minlat+$i/$binnumx*$latdiff , ",$alt ", 
		   "\n");
		print $fh "          </coordinates>
        </LinearRing>
      </outerBoundaryIs>
    </Polygon>
  </Placemark>\n";
	    }
	    print int(9*$spreadresults->[$i][$j]/$maxcount) if ($dump);
# 	    $image->set_color(127,hsv_to_rgb((($maxcount == 0) ? 0 : .8*$spreadresults->[$i][$j] / $maxcount)));
# 	    $image->draw_point($i,$j);
# 	    $image->image_orientate(3);
	}
	print "\n" if ($dump);
    }
    print $fh "</Folder>\n";
#    $image->save("tmp.jpg");
}

sub get_tomtom_style {
    # this matches the code in scripts/devices/tomtom too
    my $cache = $_[0];

    # XXX: infield vs solved...

    # traditionals break down by size
    if ($_[0]{'data'}{'subtype'} eq 'Traditional Cache') {
	if (exists($containers{$_[0]{data}{groundspeak_container}})) {
	    my $diff = $_[0]{data}{groundspeak_difficulty};
	    $diff =~ s/\.//;
	    return "#$containers{$_[0]{data}{groundspeak_container}}$diff";
	}
    } else {
	if (exists($types{$_[0]{data}{subtype}})) {
	    return "#$types{$_[0]{data}{subtype}}$_[0]{data}{groundspeak_difficulty}";
	}
	# the rest get icons for the type
    }
    return "#khStyle652";
}

1;

=head1 NAME

Kml - export kml file for Google Earth

=head1 DESCRIPTION

START HELP

  > geoqo -s 'set:import:near me' -e fortheearth.kml

The software will also produce a density plot as well, which shows the
density of caches in particular areas color-coded by relief colors
(rainbow; red = very few, purple = many).  Examples of this can be
seen on Yamar's cache homepage page http://geocaching.sharedlists.com/.

Options:

=over

=item size=NUM

  > geoqo -s 'set:import:near me' -e kml:size=100:fortheearth.kml

Changes the number of squares in the density plot to NUM by NUM.
Defaults to 50.

=item width=NUM

=item height=NUM

Changes just the width and height of the squares.  The size parameter
can change both at once.

=back

(there are a bunch of options for the density plot, which are
currently undocumented).

END HELP

=head1 AUTHOR

Wes Hardaker (AKA Yamar) <hardaker ATAT users.sourceforge.net> 

=cut
