# Copyright (C) 2007-2009 Hardaker
# License: GNU GPLv2.  See the COPYING file for details.
package GeoDB::Export::System;

use strict;
use GeoDB::Export;
use GeoDB::Utils;

our $VERSION = "1.11";

our @ISA = qw(GeoDB::Export);

our @GEOOPTIONS = 
  ({ name => 'cmd',
     help => 'COMMAND is what gets run.  If it contains a %{file:TYPE} string in it, that will be replaced with a file name of that type and it\'ll get run just once.  If it doesn\'t, it\'ll get run once per cache in the set being processed and each %{token} specifier will get replaced by that item from the cache.'
   },
   { name => 'sleep',
     help => 'Between each command executed in the per-cache mode, geoqo will sleep for NUM seconds before running the next.  It defaults to 1, and can be set to 0 to just fire them off as fast as possible.'
   },
  { name => 'dontdelete',
    help => 'if set to 1, don\'t delete the temporary files',
    nogui => 1,
  },
   { name => 'tmpdir',
     help => 'The temporary directory to use',
     default => $ENV{'HOME'} . "/.geoqo/temp",
     nogui => 1,
   }
  );

sub export_it {
    my ($self, $file, $set) = @_;

    my $cmd = $self->{'options'}{'cmd'};

    $set->load_parts({attributes => 1});

    if ($cmd =~ s/\%{file:([^\}]+)}/\%s/) {
	my $type = $1;
	my $tmpdir = $self->{'options'}{'tmpdir'};
	require File::Temp;
	import File::Temp qw/tempfile/;
	$self->{options}{'suffix'} ||= ".$type";
	$self->{options}{'suffix'} =~ s/:.*//;
	my ($fh, $filename) = tempfile("geoqoXXXXXX", 
				       DIR => $tmpdir,
				       SUFFIX => $self->{options}{'suffix'});
	$fh->close();
	my $engine = $self->new_engine($type, "Export");
	GEOERR("can't load export engine of type $type when trying to create a file for a system call\n") if (!$engine);
	GEODEBUG(5, "exporting $type to $filename\n");
	$engine->export_it("$filename", $set);
        # call system
        GEODEBUG(5, "running $cmd\n");
	system(sprintf($cmd, $filename));
	GEODEBUG(6, "return status: $?\n");
	unlink($filename) unless ($self->{'options'}{'dontdelete'});	
        return;
    }

    # run a command on each cache
    $self->{'options'}{'sleep'} = 1 if (!defined($self->{'options'}{'sleep'}));
    $set->foreach( sub {
        my $cache = shift;
        my $docmd = $cache->replace_text($cmd);
        GEODEBUG(5, "running $docmd\n");
	if (!$self->{'options'}{'dryrun'}) {
	    system($docmd);
	    GEOERR("system command failed: status=$?, error=$!\n") if ($?);
	}
        sleep($self->{'options'}{'sleep'}) if($self->{'options'}{'sleep'});
    });
}

1;

=head1 NAME

System - run system commands on each cache or on a file

=head1 DESCRIPTION

START HELP

There are two ways to use the I<system> export mechanism:

=over

=item Run a system command for cache:

  > geoqo -s 'cache:name=cool cache' -e 'system:cmd=echo %{ident} %{desc}'

That command runs echo for each cache in the set, replacing %{ident}
and %{desc} with the fields from those caches.

Useful things to do for this would be to open firefox (a web browser)
on each cache page:

  > geoqo -s 'cache:name=cool cache' -e 'system:cmd=firefox %{url}'

=item Run a system command

You can also run a command on a file generated by an export module.
As an example, the following command will call gpsbabel to convert a
gpx file to a tomtom compatible file:

  > geoqo -s 'cache:name=cool cache' -e 'system:cmd=gpsbabel -i gpx -f %{file:gpx} -o tomtom -F mycaches.ov2'

=back

END HELP

=head3 Options:

The following options can be passed to the system exporter:

=over

=item cmd=COMMAND

COMMAND is what gets run.  If it contains a %{file:TYPE} string in it,
that will be replaced with a file name of that type and it'll get run
just once.  If it doesn't, it'll get run once per cache in the set
being processed and each %{token} specifier will get replaced by that
item from the cache.

=item sleep=NUM

Between each command executed in the per-cache mode, geoqo will sleep
for NUM seconds before running the next.  It defaults to 1, and can be
set to 0 to just fire them off as fast as possible.

=back

=head1 AUTHOR

Wes Hardaker (AKA Yamar) <hardaker ATAT users.sourceforge.net> 

=cut
