# Copyright (C) 2007-2009 Hardaker
# License: GNU GPLv2.  See the COPYING file for details.
package GeoDB::Import::Loc;

require GeoDB::Import;
use GeoDB::Utils;
use GeoDB::Set;
use GeoDB::Waypoint;
use XML::Simple;
use strict;

our $VERSION = "1.11";

our @ISA = qw(GeoDB::Import);

our @GEOOPTIONS = 
({ name => 'setname',
   text => 'Overide the default set name:',
   advanced => 1,
   help => 'Uses setname as the name of geoqo SET to create based on the data. Normally this is extracted from the I<name> field within the gpx file itself and prepended with "import:".  IE, if you had a pocket query named "my search" then, when you imported the gpx file into geoqo, you\'d get a new set named "import:my search".'
 });

sub import_set {
    my ($self, $file, $dbh, %opts) = @_;
    GEODEBUG(1,"reading file: $file\n");
    $XML::Simple::PREFERRED_PARSER = "XML::Parser";
    my $xdoc = XMLin($file);

    my $set = new GeoDB::Set;
    # get the name
    my $name = $self->{'options'}{'setname'} || ("import:$file");
    GEODEBUG(2,"processing file: $file into set $name\n");

    # read in each waypoint (cache)
    my $count = 0;

    # protect against singular records
    if (ref($xdoc->{'waypoint'}) eq 'HASH' &&
	exists($xdoc->{'waypoint'}{'coord'})) {
	$xdoc = { waypoint => [$xdoc->{'waypoint'}] };
    }

    # import each cache in the series
    my ($gs, $waypoint, $type, $subtype);
    foreach my $wpt (@{$xdoc->{'waypoint'}}) {
	GEODEBUG(3, "importing record #" . 
		 $count++ . ": $wpt->{'name'}{'id'}\n");

	# these names need to be carefully matched.  we're translating
	# names in the Geocache.pm definition to the names in the XML,
	# which are almost equal in most tags except for a few differences.
	($type, $subtype) = ($wpt->{'type'} =~ /(.*)\|(.*)/);
	($type, $subtype) = ($wpt->{'type'}, '') if (!$type);
	my $owner_name;
	if ($type eq 'Waymark' &&
	    $wpt->{'name'}{'content'} =~ s/by (\w+)$//) {
	    $owner_name = $1;
	}
	$waypoint =
	  new GeoDB::Waypoint
	    ({
	     data => 
	     {
	      ident =>                         "$wpt->{'name'}{'id'}",
	      'desc' => 	     	       "$wpt->{'name'}{'content'}",
	      'lat' => 	     		       "$wpt->{'coord'}{'lat'}",
	      'lon' => 	     		       "$wpt->{'coord'}{'lon'}",
	      'type' => 	     	       "$type",
	      'subtype' => 	     	       "$subtype",
	      'url' => 	     		       "$wpt->{'link'}{'content'}",
	      'urlname' =>     		       "$wpt->{'link'}{'text'}",
	      'waymark_owner_name' =>          $owner_name,
	     }});
	$waypoint->morph_to_specific();
	$set->add($waypoint);
    }
    $set->set_name($name) if ($name);
    return $set;
}

1;

=head1 NAME

Loc - import loc files

=head1 DESCRIPTION

START HELP

  > geoqo -i myfile.loc

  > geoqo -i loc:setname=myname:myfile.loc

The loc importer imports loc files taken from the geocaching.com and
waymarking.com site (and probably many other sites too, but only
geocaching.com and waymarking.com LOC files have been tested so for)
and imports the data into your geoqo database.

END HELP

=head3 Options:

=over

=item setname=NAME

Uses setname as the name of geoqo SET to create based on the data.
Normally this is the file prepended with "import:".  IE, if you had
downloaded a .loc file and named it "myloc.loc" then when you imported
the loc file into geoqo you'd get a new set named "import:myloc.loc".

=back

=head1 AUTHOR

Wes Hardaker (AKA Yamar) <hardaker ATAT users.sourceforge.net> 

=cut
