# Copyright (C) 2007-2009 Hardaker
# License: GNU GPLv2.  See the COPYING file for details.
package GeoDB::Modify::Track;

use strict;
use GeoDB::Modify;
use GeoDB::Utils;
use XML::Simple;
use Date::Parse;

our @ISA = qw(GeoDB::Modify);

our @GEOOPTIONS = (
		   { name => 'trackfile',
		     help => 'Track data to sort by',
		   },
		   { name => 'sort',
		     default => 'trkptepoch',
		     help => 'field to sort resulting set by (if set)',
		   },
		   { name => 'distance',
		     default => '528',
		     help => 'Distance away in feet to use for determining how long you were in the area',
		   },
		   { name => 'hroffset',
		     help => 'Offset in hours to use from the datestamp found in the track; useful if the timezone in the GPS/track does not match the track was laid',
		   },
		  );

our $VERSION = "1.11";

sub calculate_timeat {

    my ($cache, $lasttime, $starttime) = @_;

    # calculate hh:mm:s we were there
    my $time = str2time($lasttime) - $starttime;
    my $sec = $time % 60;
    $time = ($time - $sec)/60;
    my $min = $time % 60;
    $time = ($time - $min)/60;

    # set start and stop times
    my @timedata = localtime($starttime);
    $cache->{'attributes'}{'trkptstart'} =
      sprintf("%02d:%02d:%02d", $timedata[2],
	      $timedata[1], $timedata[0]);

    @timedata = localtime(str2time($lasttime));
    $cache->{'attributes'}{'trkptend'} =
      sprintf("%02d:%02d:%02d", $timedata[2],
	      $timedata[1], $timedata[0]);

    $cache->{'attributes'}{'trkptnearby'} =
      sprintf("%02d:%02d:%02d", $time, $min, $sec);
}


sub modify_set {
    my ($self, $set) = @_;

    $set->load_parts({attributes => 1});

    # load the track data
    my $xdoc = XMLin($self->{'options'}{'trackfile'},
		     ForceArray => [qw(trkpt trk)],
		     KeyAttr => 'xxx');

    # pass 1: find the track point closest to each cache
    my %trackpts;
    $set->foreach(sub {
		      my $distance;
		      my $lastdistance = 10000;
		      my $cache = shift;
		      my $starttime = 0;
		      my $lasttime;

		      # XXX: this should be an array in order...
		      foreach my $track (@{$xdoc->{'trk'}}) {
			  my $count = 0;
			  next if (!exists($track->{'trkseg'}{'trkpt'}));
			  foreach my $pt (@{$track->{'trkseg'}{'trkpt'}}) {
			      $count++;
			      $distance = calc_distance($pt->{'lat'},
							$pt->{'lon'},
							$cache->{'data'}{'lat'},
							$cache->{'data'}{'lon'}
						       );
			      if ($distance < $lastdistance) {

				  # set the track source info
				  $cache->{'attributes'}{'trkptname'} =
				    $track->{'name'};
				  $cache->{'attributes'}{'trkptnum'} = $count;

				  # distance
				  $cache->{'attributes'}{'trkptdist'} =
				    $distance * 5280;

				  # time nearest data
				  $cache->{'attributes'}{'trkptepoch'} =
				    str2time($pt->{'time'});
				  if ($self->{'options'}{'hroffset'}) {
				      $cache->{'attributes'}{'trkptepoch'} +=
					$self->{'options'}{'hroffset'}*60*60;
				  }
				  my @ldata =
				    localtime($cache->{'attributes'}{'trkptepoch'});
				  $cache->{'attributes'}{'trkptstamp'} =
				    localtime($cache->{'attributes'}{'trkptepoch'});
				  $cache->{'attributes'}{'trkptdate'} =
				    sprintf("%04d-%02d-%02d", $ldata[5]+1900,
					    $ldata[4]+1, $ldata[$3]);
				  $cache->{'attributes'}{'trkpttime'} =
				    sprintf("%02d:%02d:%02d", $ldata[2],
					    $ldata[1], $ldata[0]);

				  # reset the nearby field just in case
#				  $cache->{'attributes'}{'trkptnearby'} = 0;

				  # remember the distance
				  $lastdistance = $distance;
			      }

			      # remember start time if we first get close
			      if (!$starttime &&
				  $distance*5280 < $self->{'options'}{'distance'}) {
				  $starttime = str2time($pt->{time});
#				  print "here1: $cache->{'attributes'}{'ident'} $starttime && " . $distance*5280 . " < $self->{'options'}{'distance'}\n";
			      }

			      # then when we move away again, generate
			      # more data based on how long we were
			      # there.
			      if ($distance*5280 > $self->{'options'}{'distance'} &&
				  $starttime && !$cache->{'attributes'}{'trkptnearby'}) {

				  calculate_timeat($cache,
						   $lasttime, $starttime);
			      }
			      $lasttime = $pt->{time};
			  }

			  # track ended...  assume we turned off near
			  # the cache (or else some GPSs start with a
			  # time much later but the older lat/lon till
			  # it syncs and you get really bad data
			  if ($starttime && !$cache->{'attributes'}{'trkptnearby'}) {
			      calculate_timeat($cache,
					       $lasttime, $starttime);
			  }
		      }
		  });
	
    # pass 2: do the sorting
    if ($self->{'options'}{'sort'}) {
	$set->sortby(sub ($$) { 
			 $_[0]->{'attributes'}{$self->{'options'}{'sort'}} <=>
			   $_[1]->{'attributes'}{$self->{'options'}{'sort'}}
		     });
    }
}

1;

=head1 NAME

Track - Adds data to the found waypoints based on information from a track

=head1 DESCRIPTION

START HELP

The track plugin lets you compare a list of waypoints found in a search
against the data in a gpx track file.  Specifically, it:

 - adds the following attribute fields to the waypoint's attribuets
   (not permenantly; use --save to save them):

   trkptdate    -  The date of the track point nearest the cache
   trkpttime    -  The time of the track point nearest the cache
   trkptnum     -  The number of the track point nearest the cache
   trkptname    -  The name of the track the point is in

   trkptdist    -  Distance in feet from the cache to the nearest track point
   trkptnum     -  The Nth track point counter into the track point data
   trkptepoch   -  Seconds since 1970 of the trackpoint closest to the cache
   trkptstart   -  Start time of the search (got closer than config distance)
   rtkptend     -  End time of the search
   rtkptnearby  -  Amount of time spent searching (= end - start)

 - optionally sorts the resulting waypoints by time visited if the
   sort option is set.

Using this and the track search plugin you can take a track GPX file
downloaded from a GPS and list all the caches you came near in the
order you went to them along with the time stamp when you visited it.
As an example, this is the results of a trip I took one day:

  > geoqo -s 'track:track.gpx' -m 'track:track.gpx' -d list:'%{7.7:ident} %{attr:trkpttime}  %{urlname}'
  
  GC10J27 10:55:22  This one is Golden
  GCPC2R  11:16:53  "THE ISLAND"
  GCYCGE  13:02:07  Legacy Trail #1 - River BBQ
  GCD8A1  13:10:55  Under Lock & Key
  GC5CBC  13:52:32  Let the Games Begin
  GCPR6E  14:11:27  YFcache10
  GCWR36  14:31:10  Granite Rock
  GCK8YR  14:54:07  The Cache in The Middle of The Freeway
  GCYYXV  16:02:05  Cisco Grove TB Stop
  GC10Q0Q 16:02:41  Hard Rock Exchange #2
  GCW1C3  16:27:08  Crack with a View
  GCG38E  16:45:33  The Solar System - Pluto

To get track data from the gps, use gpsbabel:

  gpsbabel -t -i garmin -f usb: -o gpx -F track.gpx

END HELP

=head1 AUTHOR

Wes Hardaker (AKA Yamar) <hardaker ATAT users.sourceforge.net>

=cut
